#include <cmath>
#include "Log.hpp"
#include "Renderer.hpp"
#include "Alermath.hpp"
#include "SFMLOrthogonalLayer.hpp"
#include "Spectre.hpp"

#define WIDTH 1280
#define HEIGHT 720

Renderer* Renderer::s_self = nullptr;

Renderer::Renderer(sf::RenderWindow* window):
m_window(window) {
	m_view.setSize(WIDTH, HEIGHT);
	s_self = this;

	

	Log(INFO, "Renderer up.");
}

void Renderer::registerDrawable(sf::Drawable* drawable, unsigned z_index, Mode mode, std::string name) {
	if (!s_self) {
		Log(ERROR, "Trying to register a drawable without a renderer.");
	}
	RenderObject ro;
	ro.drawable = drawable;
	ro.z_index = z_index;
	ro.mode = mode;
	ro.name = name;
	s_self->m_data.push_back(ro);

	Log(INFO, "Registered drawable ", name);
}

void Renderer::deregisterDrawable(sf::Drawable* drawable) {
	if (!s_self) {
		Log(ERROR, "Trying to deregister a drawable without a renderer.");
	}
	
	Log(INFO, "Removing");
	list();
	s_self->m_data.erase(std::remove_if(s_self->m_data.begin(), s_self->m_data.end(), [drawable](auto pair){
		if (pair.drawable == drawable) {
			Log(INFO, pair.name);
		}
		return pair.drawable == drawable;
	}), s_self->m_data.end());
	list();
	Log(INFO, "Deregistered drawable ");
}

void Renderer::registerLayer(MapLayer* layer) {
	if (!s_self) {
		Log(ERROR, "Trying to register a layer without a renderer.");
	}

	s_self->m_layers.push_back(layer);
	layer->updateVisibility(s_self->m_view);

	Log(INFO, "Registered layer");
}
void Renderer::deregisterLayer(MapLayer* layer) {
	if (!s_self) {
		Log(ERROR, "Trying to deregister a layer without a renderer.");
	}

	s_self->m_layers.erase(std::remove_if(s_self->m_layers.begin(), s_self->m_layers.end(), [layer](auto pair) {
		return pair == layer;
		}), s_self->m_layers.end());

	Log(INFO, "Deregistered layer");
}

void Renderer::attach(sf::Vector2f pos) {
	s_self->m_view.setCenter(pos);
	for (const MapLayer* layer : s_self->m_layers) {
		layer->updateVisibility(s_self->m_view);
	}
}

void Renderer::follow(sf::Transformable* followed) {
	s_self->m_followed = followed;
}

sf::Vector2f Renderer::getPosition() {
	if (s_self->m_followed) {
		return s_self->m_followed->getPosition();
	}
	else {
		return sf::Vector2f(0.f, 0.f);
	}
}

void Renderer::sort() {
	std::sort(s_self->m_data.begin(), s_self->m_data.end(), [](auto o1, auto o2)->bool {
		return o1.mode == o2.mode ? o1.z_index < o2.z_index : o1.mode < o2.mode;
	});

	Log(INFO, "Renderer sorted");
}

void Renderer::clear() {
	m_data.clear();
}

void Renderer::update(float dt) {
	float x1 = m_view.getCenter().x;
	float x2 = getPosition().x;
	float y1 = m_view.getCenter().y;
	float y2 = getPosition().y;

	const float camera_speed = (400.f * TOOFAST);// -(Spectre::getChaos() < 0.5 ? Spectre::getChaos() * 2.f : 1.f) * 50.f;

	if(x1 != x2 || y1 != y2) {
		if(std::fabs(x2 - x1) > dt * camera_speed) {
			m_view.move(dt * camera_speed * math::sgn(x2 - x1), 0);
		} else {
			m_view.move(x2 - x1, 0);
		}
		if (std::fabs(y2 - y1) > dt * camera_speed) {
			m_view.move(0, dt * camera_speed * math::sgn(y2 - y1));
		}
		else {
			m_view.move(0, y2 - y1);
		}
		float shake_strength = (Spectre::getChaos() - CHAOS_CAMERASHAKE_THRESHOLD) / (1.f - CHAOS_CAMERASHAKE_THRESHOLD);
		if (Spectre::getChaos() >= CHAOS_CAMERASHAKE_THRESHOLD && rand() % 100 < 100.f * shake_strength) {
			m_view.move(dt * (90.f + (50.f * shake_strength)) * (rand() % 2 ? -1.f : 1.f), dt * (90.f + (50.f * shake_strength)) * (rand() % 2 ? -1.f : 1.f));
		}
		for (const MapLayer* layer : m_layers) {
			layer->updateVisibility(s_self->m_view);
		}
	}

	if (m_whack) {
		static float elapsed = 0.f;
		static sf::Clock rotate_clock;
		static float rotate_count = 0.6f;
		elapsed += dt;
		float destw = WIDTH - elapsed * elapsed * float(WIDTH / 30.f);
		float desth = HEIGHT - elapsed * elapsed * float(HEIGHT / 30.f) * 1.2f;

		if (destw < 0 || desth < 0) {
			m_dead = true;
			return;
		}
		m_view.setSize(destw, desth);
		if (rotate_clock.getElapsedTime().asSeconds() > 0.15f + math::non_negative(rotate_count)) {
			m_view.setRotation(rand() % 360);
			rotate_clock.restart();
			rotate_count -= 0.1f;
			rotate_count /= 1.5f;
		}
	}
}

void Renderer::list() {
	std::string list;
	for (const RenderObject& ro : s_self->m_data) {
		list += ro.name + " ";
	}
	Log(INFO, "Render list: ", list);
}

void Renderer::render() {
	bool default_view = true;
	m_window->setView(m_window->getDefaultView());

	for(RenderObject& ro: m_data) {
		if (ro.mode == BG && !default_view) {
			default_view = true;
			m_window->setView(m_window->getDefaultView());
		}
		if(ro.mode == FG && default_view) {
			default_view = false;
			m_window->setView(m_view);
		}
		if(ro.mode == UI && !default_view) {
			default_view = true;
			m_window->setView(m_window->getDefaultView());
		}
		// Log(INFO, ro.name);
		m_window->draw(*ro.drawable);
	}
	m_window->display();
}